from dataclasses import dataclass
from typing import Optional, List


@dataclass
class ExperimentVariable:
    """Information about a experimental variable

    The number of occurrences of :class:`~ExperimentVariable` is specified by
    the value of :attr:`~VamasHeader.num_experiment_variables` in
    :class:`~VamasHeader`.

    Attributes:
        label (str):
        unit (str):
            ( c/s | d | degree | eV | K | micro C | micro m | m/s | n | nA
            | ps | s | u | V )
            These values are abbreviations for the units listed below:

            +---------+-------------------------------------------------------+
            | c/s     | counts per second                                     |
            +---------+-------------------------------------------------------+
            | d       | dimensionless - just a number, eg. counts per channel |
            +---------+-------------------------------------------------------+
            | degree  | angle in degrees                                      |
            +---------+-------------------------------------------------------+
            | eV      | electron volts                                        |
            +---------+-------------------------------------------------------+
            | K       | Kelvin                                                |
            +---------+-------------------------------------------------------+
            | micro C | microcoulombs                                         |
            +---------+-------------------------------------------------------+
            | micro m | micrometres                                           |
            +---------+-------------------------------------------------------+
            | m/s     | metres per second                                     |
            +---------+-------------------------------------------------------+
            | n       | not defined here - may be given in a label            |
            +---------+-------------------------------------------------------+
            | nA      | nanoamps                                              |
            +---------+-------------------------------------------------------+
            | ps      | picoseconds                                           |
            +---------+-------------------------------------------------------+
            | s       | seconds                                               |
            +---------+-------------------------------------------------------+
            | u       | unified atomic mass units                             |
            +---------+-------------------------------------------------------+
            | V       | volts                                                 |
            +---------+-------------------------------------------------------+

    """

    label: str
    unit: str


@dataclass
class FutureUpgradeExperimentEntry:
    """Information about future uprade experiment entries

    The number of occurrences of :class:`~FutureUpgradeExperimentEntry` is
    given by the value of
    :attr:`~VamasHeader.num_future_upgrade_experiment_entries` in
    :class:`~VamasHeader`.
    It is defined as a text line so that any integer, real number or text line
    inserted here by a future upgrade of the format can be read as a text line
    then discarded.

    Attributes:
        label (str):
        unit (str):

    """

    label: str
    unit: str


@dataclass
class VamasHeader:
    """Header information about a Vamas experiment



    Attributes:
        format_identifier (str): Format identifier.
            Must be 'VAMAS Surface Chemical Analysis Standard Data Transfer
            Format 1988 May 4'.
        institution_identifier (str): Institution identifier.
        instrument_model_identifier (str): Instrumental Model Identifier.
        operator_identifier (str): Operator Identifier.
        experiment_identifier (str): Experiment Identifier.
        num_lines_comment (int): Number of comment lines.
        comment (str): Concatenated comment lines.
            The comment may include details of the last calibration
            of the instrument.
        experiment_mode (str): Experiment mode.
            Has one of the values 'MAP' | 'MAPDP' | 'MAPSV' | 'MAPSVDP'
            | 'NORM' | 'SDP' | 'SDPSV' | 'SEM'.
            The contents of each block in the experiment are indicated by the
            values of experiment mode as follows:

            - MAP
                A spectrum which refers to a specified point in a
                regular two-dimensional spatial array.
            - MAPDP
                A spectrum which refers to a specified point in a regular
                two-dimensional spatial array and to a specified layer in a
                depth profile.
            - MAPSV
                A complete set of single values of a fixed number of
                variables for every point in a regular two-dimensional spatial
                array. Note that an x linescan consists of a map with the value
                of number of analysis positions equal to the value of number of
                discrete x coordinates available in full map, that is, the
                number of discrete y coordinates is unity; in a y linescan the
                roles of x and y are reversed.
            - MAPSVDP
                A complete set of single values of a fixed number of
                variables for every point in a regular two-dimensional array
                for one layer in a depth profile. Successive blocks refer to
                successive layers in the depth profile.
            - NORM
                Either independent data or data which refers to a
                specified set of single values of one or more experimental
                variables; the data may be spectral or non-spectral.
            - SDP
                A spectrum which refers to a specified layer in a depth
                profile.
            - SDPSV
                A complete set of single values of a fixed number of
                variables for every layer in a depth profile.
            - SEM
                An electron emission intensity for every point in a
                regular two-dimensional spatial array.
        scan_mode (str): Scan Mode.
            Has one of the values 'REGULAR' | 'IRREGULAR' | 'MAPPING'.
            If the value of :attr:`~VamasHeader.experiment_mode` is **MAPSV**,
            **MAPSVDP** or **SEM** then the value of
            :attr:`~VamasHeader.scan_mode` must be **MAPPING**, otherwise if
            the data is in the form of an abscissa start, an abscissa increment
            and a number of complete sets of values of one or more experimental
            variables then the value of :attr:`~VamasHeader.scan_mode` is
            **REGULAR**, otherwise the value of :attr:`~VamasHeader.scan_mode`
            is **IRREGULAR**.
        num_spectral_regions (int): Number of spectral regions.
            Normally only one technique is used in an experiment but there may
            be more. The value of :attr:`~VamasHeader.num_spectral_regions` is
            the sum for all techniques of the numbers of spectral regions in
            each technique.
            :attr:`~VamasHeader.num_spectral_regions` is inserted if and only
            if the value of :attr:`~VamasHeader.experiment_mode` is **MAP**,
            **MAPDP**, **NORM** or **SDP**. (optional-sequence)
        num_analysis_positions (int): Number of analysis positions.
            Inserted if and only if the value of
            :attr:`~VamasHeader.experiment_mode` is either **MAP** or
            **MAPDP**. (optional-sequence)
        num_discrete_x_coords_in_full_map (int): Number of discrete x
            coordinates available in full map.
            Inserted if and only if the value of
            :attr:`~VamasHeader.experiment_mode` is either **MAP** or
            **MAPDP**. (optional-sequence)
        num_discrete_y_coords_in_full_map (int): Number of discrete y
            coordinates available in full map.
            Inserted if and only if the value of
            :attr:`~VamasHeader.experiment_mode` is either **MAP** or
            **MAPDP**. (optional-sequence)
        num_experiment_variables (int): Number of experimental variables
            An experimental variable is a parameter which may be varied from
            block to block through the experiment but which remains constant
            within each block.
        experiment_variables (Optional[List[ExperimentVariable]]): Experiment
            variables. (optional-sequence)

        num_entries_inclusion_exclusion (int): Parameter inclusion or
            exclusion. (prefix number)
        block_params_includes (List[bool]): Deteminated which parameters are
            included in all blocks.
        num_manually_entered_items_in_block (int): Prefix number of manually
            entered item.
            The number of occurrences of *prefix number of manually entered
            item* is specified by the value of
            :attr:`~VamasHeader.num_manually_entered_items_in_block` above.
            If this is greater than zero then the values of successive
            occurrences of *prefix number of manually entered item* should
            be in ascending order. Any of the items preceded by prefix numbers
            in comment brackets in the syntax-rule defining *block* which need
            to be evaluated by the operator and manually entered from the
            keyboard should be included in this list. If an item is to be
            expressed as a real number and the operator is unable to supply a
            value then the computer should enter the value 1E37.
            (repeated-sequence)
        num_future_upgrade_experiment_entries (int): Number of future upgrade
            experiment entries.
        future_upgrade_experiment_entries
            (Optional[List[FutureUpgradeExperimentEntry]]):
            List of future upgrade experiment entries.
            The number of occurrences of :class:`~FutureUpgradeExperimentEntry`
            is given by the value of
            :attr:`~VamasHeader.num_future_upgrade_experiment_entries`.
            It is defined as a text line so that any integer, real number or
            text line inserted here by a future upgrade of the format can be
            read as a text line then discarded. (repeated-sequence)
        num_future_upgrade_block_entries (int):
            :attr:`~VamasHeader.num_future_upgrade_experiment_entries` and
            :attr:`~VamasHeader.num_future_upgrade_block_entries` are included
            in case the format is upgraded in the future to include more
            non-optional, non-repeating parameters. The numbers of these new
            parameters will be entered here so that old programs can skip the
            new parameters in new data, and new programs will not try to read
            the new parameters in old data. For the present both of them would
            be set to zero.
        num_blocks (int): Number of blocks.

    """

    format_identifier: str
    institution_identifier: str
    instrument_model_identifier: str
    operator_identifier: str
    experiment_identifier: str
    num_lines_comment: int
    comment: Optional[str]
    experiment_mode: str
    scan_mode: str
    num_experiment_variables: int
    num_entries_inclusion_exclusion: int
    block_params_includes: List[bool]
    num_manually_entered_items_in_block: int
    num_future_upgrade_experiment_entries: int
    num_future_upgrade_block_entries: int
    num_blocks: int

    # Optional-Sequences
    num_spectral_regions: Optional[int] = None
    num_analysis_positions: Optional[int] = None
    num_discrete_x_coords_in_full_map: Optional[int] = None
    num_discrete_y_coords_in_full_map: Optional[int] = None
    experiment_variables: Optional[List[ExperimentVariable]] = None
    future_upgrade_experiment_entries: Optional[
        List[FutureUpgradeExperimentEntry]
    ] = None



from typing import Union, List, Dict, Tuple, TextIO
from pathlib import Path

from .vamas_header import (
    VamasHeader,
    ExperimentVariable,
    FutureUpgradeExperimentEntry,
)
from .vamas_block import (
    VamasBlock,
    SputteringSource,
    LinescanCoordinates,
    CorrespondingVariable,
    AdditionalNumericalParam,
)

from .errors import VmsIdentifierError, FileExtensionError


class Vamas:
    """Main class for handling a vamas file

    Parses the vamas file into the attributes header and blocks.

    Args:
        file (Union[str, Path]): vamas file to be parsed

    Attributes:
        header (VamasHeader):
        blocks (List[VamasBlock]):
    """

    def __init__(self, file: Union[str, Path]) -> None:
        if not str(file).endswith(".vms"):
            raise FileExtensionError

        with open(file) as f:
            self.header, self.blocks = _read_vamas(f)


def _read_vamas(f: TextIO) -> Tuple[VamasHeader, List[VamasBlock]]:
    """Parses a vamas file

    Args:
        f (TextIO): file descripter for a vamas file


    Returns:
        Parsed vamas-file as tuple of :class:`~vamas.vamas_header.VamasHeader`
        a list of :class:`~vamas.vamas_block.VamasBlock`
    """

    h: Dict = {}
    h["format_identifier"] = next(f).strip()

    if (
        h["format_identifier"]
        != "VAMAS Surface Chemical Analysis Standard Data Transfer Format"
        " 1988 May 4"
    ):
        raise VmsIdentifierError

    h["institution_identifier"] = next(f).strip()
    h["instrument_model_identifier"] = next(f).strip()
    h["operator_identifier"] = next(f).strip()
    h["experiment_identifier"] = next(f).strip()

    h["num_lines_comment"] = int(next(f))
    comments = []
    for _ in range(h["num_lines_comment"]):
        comments.append(next(f).strip())
    h["comment"] = "\n".join(comments)

    h["experiment_mode"] = next(f).strip()
    h["scan_mode"] = next(f).strip()

    if h["experiment_mode"] in ["MAP", "MAPD", "NORM", "SDP"]:
        h["num_spectral_regions"] = int(next(f).strip())

    if h["experiment_mode"] in ["MAP", "MAPD"]:
        h["num_analysis_positions"] = int(next(f))
        h["num_discrete_x_coords_in_full_map"] = int(next(f))
        h["num_discrete_y_coords_in_full_map"] = int(next(f))

    h["num_experiment_variables"] = int(next(f))
    h["experiment_variables"] = []
    for _ in range(h["num_experiment_variables"]):
        h["experiment_variables"].append(
            # {"label": next(f).strip(), "unit": next(f).strip()}
            ExperimentVariable(next(f).strip(), next(f).strip())
        )

    h["num_entries_inclusion_exclusion"] = int(next(f))
    h["block_params_includes"] = [
        h["num_entries_inclusion_exclusion"] <= 0 for _ in range(40)
    ]
    for _ in range(abs(h["num_entries_inclusion_exclusion"])):
        h["block_params_includes"][int(next(f)) + 1] = (
            h["num_entries_inclusion_exclusion"] > 0
        )

    h["num_manually_entered_items_in_block"] = int(next(f))

    h["num_future_upgrade_experiment_entries"] = int(next(f))
    h["future_upgrade_experiment_entries"] = []
    for _ in range(h["num_future_upgrade_experiment_entries"]):
        h["future_upgrade_experiment_entries"].append(
            # {"label": next(f).strip(), "unit": next(f).strip()}
            FutureUpgradeExperimentEntry(next(f).strip(), next(f).strip())
        )

    if h["num_future_upgrade_experiment_entries"] != 0:
        print("unsupported future upgrade experiment entries")

    h["num_future_upgrade_block_entries"] = int(next(f))
    if h["num_future_upgrade_block_entries"] != 0:
        print("unsupported future block entries")

    h["num_blocks"] = int(next(f))

    # End of header
    blocks: List[Dict] = []
    for _ in range(h["num_blocks"]):
        include = (
            [True for _ in range(40)]
            if len(blocks) == 0
            else h["block_params_includes"]
        )
        # b = VamasBlock()
        # fb = VamasBlock()
        b: Dict = {}
        fb: Dict = {}

        b["block_identifier"] = next(f).strip()
        b["sample_identifier"] = next(f).strip()

        b["year"] = int(next(f)) if include[0] else fb["year"]
        b["month"] = int(next(f)) if include[1] else fb["month"]
        b["day"] = int(next(f)) if include[2] else fb["day"]
        b["hour"] = int(next(f)) if include[3] else fb["hour"]
        b["minute"] = int(next(f)) if include[4] else fb["minute"]
        b["second"] = int(next(f)) if include[5] else fb["second"]

        b["num_hours_advance_gmt"] = (
            float(next(f)) if include[6] else fb["num_hours_advance_gmt"]
        )

        if include[7]:
            b["num_lines_block_comment"] = int(next(f))
            block_comments = []
            for _ in range(b["num_lines_block_comment"]):
                block_comments.append(next(f).strip())

            b["block_comment"] = "\n".join(block_comments)
        else:
            b["num_lines_block_comment"] = fb["num_lines_block_comment"]
            b["block_comment"] = fb["block_comment"]

        b["technique"] = next(f).strip() if include[8] else fb["technique"]

        if h["experiment_mode"] in ["MAP", "MAPD"]:
            b["x_coord"] = int(float(next(f))) if include[9] else fb["x_coord"]
            b["y_coord"] = int(float(next(f))) if include[9] else fb["y_coord"]

        if include[10]:
            b["values_exp_var"] = []
            for _ in range(len(h["experiment_variables"])):
                b["values_exp_var"].append(next(f))
        else:
            b["values_exp_var"] = fb["values_exp_var"]

        b["analysis_source_label"] = (
            next(f).strip() if include[11] else fb["analysis_source_label"]
        )

        if h["experiment_mode"] in ["MAPDP", "MAPSVDP", "SDP", "SDPSV"] or b[
            "technique"
        ] in [
            "SNMS energy spec",
            "FABMS",
            "FABMS energy spec",
            "ISS",
            "SIMS",
            "SIMS energy spec",
            "SNMS",
        ]:
            b["sputtering_z"] = (
                int(next(f)) if include[12] else fb["sputtering_z"]
            )
            b["sputtering_num_particles"] = (
                float(next(f))
                if include[12]
                else fb["sputtering_num_particles"]
            )
            b["sputtering_charge"] = (
                float(next(f).strip())
                if include[12]
                else fb["sputtering_charge"]
            )

        b["analysis_source_characteristic_energy"] = (
            float(next(f))
            if include[13]
            else fb["analysis_source_characteristic_energy"]
        )
        b["analysis_source_strength"] = (
            float(next(f)) if include[14] else fb["analysis_source_strength"]
        )
        b["analysis_source_beam_width_x"] = (
            float(next(f))
            if include[15]
            else fb["analysis_source_beam_width_x"]
        )
        b["analysis_source_beam_width_y"] = (
            float(next(f))
            if include[15]
            else fb["analysis_source_beam_width_y"]
        )

        if h["experiment_mode"] in [
            "MAP",
            "MAPDP",
            "MAPSV",
            "MAPSVDP",
            "SEM",
        ]:
            b["field_view_x"] = (
                float(next(f)) if include[16] else fb["field_view_x"]
            )
            b["field_view_y"] = (
                float(next(f)) if include[16] else fb["field_view_y"]
            )

        if h["experiment_mode"] in ["MAPSV", "MAPSVDP", "SEM"]:
            if include[17]:
                b["linescan_coordinates"] = LinescanCoordinates(
                    first_linescan_start_x=int(next(f)),
                    first_linescan_start_y=int(next(f)),
                    first_linescan_finish_x=int(next(f)),
                    first_linescan_finish_y=int(next(f)),
                    last_linescan_finish_x=int(next(f)),
                    last_linescan_finish_y=int(next(f)),
                )
            else:
                b["linescan_coordinates"] = fb["linescan_coordinates"]

        b["analysis_source_polar_incidence_angle"] = (
            float(next(f))
            if include[18]
            else fb["analysis_source_polar_incidence_angle"]
        )
        b["analysis_source_azimuth"] = (
            float(next(f)) if include[19] else fb["analysis_source_azimuth"]
        )
        b["analyzer_mode"] = (
            next(f).strip() if include[20] else fb["analyzer_mode"]
        )

        b["analyzer_pass_energy_or_retard_ratio_or_mass_res"] = (
            float(next(f))
            if include[21]
            else fb["analyzer_pass_energy_or_retard_ratio_or_mass_res"]
        )

        if b["technique"] == "AES diff":
            b["differential_width"] = (
                float(next(f)) if include[22] else fb["differential_width"]
            )

        b["magnification_analyzer_transfer_lens"] = (
            float(next(f))
            if include[23]
            else fb["magnification_analyzer_transfer_lens"]
        )
        b["analyzer_work_function_or_acceptance_energy"] = (
            float(next(f))
            if include[24]
            else fb["analyzer_work_function_or_acceptance_energy"]
        )

        b["target_bias"] = float(next(f)) if include[25] else fb["target_bias"]

        b["analysis_width_x"] = (
            float(next(f)) if include[26] else fb["analysis_width_x"]
        )
        b["analysis_width_y"] = (
            float(next(f)) if include[26] else fb["analysis_width_y"]
        )

        b["analyzer_axis_take_off_polar_angle"] = (
            float(next(f))
            if include[27]
            else fb["analyzer_axis_take_off_polar_angle"]
        )
        b["analyzer_axis_take_off_azimuth"] = (
            float(next(f))
            if include[27]
            else fb["analyzer_axis_take_off_azimuth"]
        )

        b["species_label"] = (
            next(f).strip() if include[28] else fb["species_label"]
        )

        b["transition_or_charge_state_label"] = (
            next(f).strip()
            if include[29]
            else fb["transition_or_charge_state_label"]
        )
        b["charge_detected_particle"] = (
            int(next(f)) if include[29] else fb["charge_detected_particle"]
        )

        if h["scan_mode"] != "REGULAR":
            print("Only REGULAR scans supported")

        b["x_label"] = next(f).strip() if include[30] else fb["x_label"]
        b["x_units"] = next(f).strip() if include[30] else fb["x_units"]

        b["x_start"] = float(next(f)) if include[30] else fb["x_start"]
        b["x_step"] = float(next(f)) if include[30] else fb["x_step"]

        if include[31]:
            b["num_corresponding_variables"] = int(next(f))
            b["corresponding_variables"] = []
            for _ in range(b["num_corresponding_variables"]):
                b["corresponding_variables"].append(
                    CorrespondingVariable(
                        label=next(f).strip(),
                        unit=next(f).strip(),
                        y_values=[],
                    )
                )

        else:
            b["num_corresponding_variables"] = fb[
                "num_corresponding_variables"
            ]
            assert fb["corresponding_variables"] is not None
            b["corresponding_variables"] = fb["corresponding_variables"].copy()

        b["signal_mode"] = (
            next(f).strip() if include[32] else fb["signal_mode"]
        )

        b["signal_collection_time"] = (
            float(next(f)) if include[33] else fb["signal_collection_time"]
        )

        b["num_scans_to_compile_block"] = (
            int(next(f)) if include[34] else fb["num_scans_to_compile_block"]
        )

        b["signal_time_correction"] = (
            float(next(f)) if include[35] else fb["signal_time_correction"]
        )

        if h["experiment_mode"] in ["MAPDP", "MAPSVDP", "SDP", "SDPSV"] and b[
            "technique"
        ] in [
            "AES diff",
            "AES dir",
            "EDX",
            "ELS",
            "UPS",
            "XRF",
        ]:
            if include[36]:
                b["sputtering_source"] = SputteringSource(
                    energy=float(next(f)),
                    beam_current=float(next(f)),
                    width_x=float(next(f)),
                    width_y=float(next(f)),
                    polar_incidence_angle=float(next(f)),
                    azimuth=float(next(f)),
                    mode=next(f).strip(),
                )
            else:
                b["sputtering_source"] = fb["sputtering_source"]

        b["sample_normal_polar_angle_tilt"] = (
            float(next(f))
            if include[37]
            else fb["sample_normal_polar_angle_tilt"]
        )

        b["sample_normal_tilt_azimuth"] = (
            float(next(f)) if include[37] else fb["sample_normal_tilt_azimuth"]
        )

        b["sample_rotation_angle"] = (
            float(next(f)) if include[38] else fb["sample_rotation_angle"]
        )

        if include[39]:
            b["num_additional_numerical_params"] = int(next(f))
            b["additional_numerical_params"] = []
            for _ in range(b["num_additional_numerical_params"]):
                b["additional_numerical_params"].append(
                    AdditionalNumericalParam(
                        label=next(f).strip(),
                        unit=next(f).strip(),
                        value=float(next(f)),
                    )
                )
        else:
            b["num_additional_numerical_params"] = fb[
                "num_additional_numerical_params"
            ]
            b["additional_numerical_params"] = fb[
                "additional_numerical_params"
            ]

        b["num_y_values"] = int(next(f))
        for corres_var in b["corresponding_variables"]:
            corres_var.y_min = float(next(f))
            corres_var.y_max = float(next(f))

        for _ in range(
            int(b["num_y_values"] / len(b["corresponding_variables"]))
        ):
            for corres_var in b["corresponding_variables"]:
                corres_var.y_values.append(float(next(f)))

        blocks.append(b)
        fb = blocks[0]

    return VamasHeader(**h), [VamasBlock(**item) for item in blocks]



from dataclasses import dataclass
from typing import Optional, List


@dataclass
class LinescanCoordinates:
    """Information about linescan coordinates in mapping experiments

    Inserted if and only if the value of
    :attr:`VamasHeader.experiment_mode
    <vamas.vamas_header.VamasHeader.experiment_mode>` is **MAPSV**,
    **MAPSVDP** or **SEM**.
    They are required for specifying the size and shape of the map and for
    relating the order in the scan sequence to the position on the sample.
    In the coordinate system to be used, x-values start at unity at the
    left-hand side of the frame and increase towards the right-hand side,
    and y-values start at unity at the top of the frame and increase towards
    the bottom of the frame, as shown below.
    **(18)**

    +------------------+
    | 1, 1        N, 1 |
    |                  |
    |                  |
    |                  |
    | 1, M        N, M |
    +------------------+

    Attributes:
        first_linescan_start_x (int): Start x-value of the first linescan.
        first_linescan_start_y (int): Start y-value of the first linescan.
        first_linescan_finish_x (int): End x-value of the first linescan.
        first_linescan_finish_y (int): End y-value of the first linescan.
        last_linescan_finish_x (int): End x-value of the last linescan.
        last_linescan_finish_y (int): End y-value of the last linescan.

    """

    first_linescan_start_x: int
    first_linescan_start_y: int
    first_linescan_finish_x: int
    first_linescan_finish_y: int
    last_linescan_finish_x: int
    last_linescan_finish_y: int


@dataclass
class SputteringSource:
    """Information about the sputtering source

    For a sputtering source used in addition to the analysis source, as in
    depth profiling.
    Inserted if and only if both (1) the value of :attr:`VamasBlock.technique`
    is **AES diff**, **AES dir**, **EDX**, **ELS**, **UPS**, **XPS** or
    **XRF**, and (2) the value of :attr:`VamasHeader.experiment_mode
    <vamas.vamas_header.VamasHeader.experiment_mode>` is **MAPDP**,
    **MAPSVDP**, **SDP** or **SDPSV**.
    **(37)**


    Attributes:
        energy (float): Energy of the sputtering source in electron volts.
        beam_current (float): Beam current of the sputtering source in nanoamps
            or equivalent for neutrals.
        width_x (float): X-width in micrometres at the sample in the plane
            perpendicular to the sputtering source beam.
        width_y (float): Y-width in micrometres at the sample in the plane
            perpendicular to the sputtering source beam.
        polar_incidence_angle (float): Polar incidence angle in degrees from
            upward z-direction, defined by the sample stage.
        azimuth (float): Azimuth angle in degrees clockwise from the
            y-direction towards the operator, defined by the sample stage
        mode (str): Mode of the sputtering source.
            The value is either 'continuous', when sputtering continues while
            spectral data is being recorded, or 'cyclic', when sputtering is
            suspended while spectral data is being recorded.

    """

    energy: float
    beam_current: float
    width_x: float
    width_y: float
    polar_incidence_angle: float
    azimuth: float
    mode: str


@dataclass
class CorrespondingVariable:
    """Information about the measured values

    The number of occurrences of :class:`CorrespondingVariable` is specified by
    the value of :attr:`VamasBlock.num_corresponding_variables`.
    The order in which the entries appear is the same as the order in
    which the corresponding values of :attr:`~CorrespondingVariable.label`
    are given.

    Attributes:
        label (str): Label of the corresponding variable.
        unit (str): Unit of the corresponding variable.
        y_values (List[float]): Measured y-values.
            The number of occurrences of ordinate values is specified by the
            value of :attr:`VamasBlock.num_y_values`. If this value is greater
            than unity then the data is sent in the form of successive
            complete sets, each set consisting of an ordinate y_value for each
            of the corresponding variables arranged in the same order as that
            in which each value of :attr:`~CorrespondingVariable.label` is
            given.
            The minus-sign followed by the empty-sequence indicates that there
            must be at least one y value.
        y_min (float): Minimal measured y-value.
            The number of occurrences of entries is specified by the value of
            :attr:`VamasBlock.num_corresponding_variables`. The order in which
            the entries appear is the same as the order in which the
            corresponding values of :attr:`~CorrespondingVariable.label`
            are given.
        y_max (float): Maximal measured y-value.
            The number of occurrences of entries is specified by the value of
            :attr:`VamasBlock.num_corresponding_variables`. The order in which
            the entries appear is the same as the order in which the
            corresponding values of :attr:`~CorrespondingVariable.label`
            are given.
    """

    label: str
    unit: str
    y_values: List[float]
    y_min: Optional[float] = None
    y_max: Optional[float] = None


@dataclass
class AdditionalNumericalParam:
    """Information about additional numerical parameters

    The number of occurrences of AdditionalNumericalParam is specified by the
    value of :attr:`VamasBlock.num_additional_numerical_params`.

    Attributes:
        label (str): Label of the additional numerical parameter.
        unit (str): Unit of the additional numerical parameter.
        value (str): Value of the additional numerical parameter.

    """

    label: str
    unit: str
    value: float


@dataclass
class VamasBlock:
    """Information about a measurement as part of an experiment

    VamasBlock contains the information about one part of the experiment, e.g.
    one spectral region or scan in XPS measurements or one image in an
    SEM experiment.

    Attributes:
        block_identifier (str): Identifier of the block.
        sample_identifier (str): Identifier of the sample.
        year (int): Year of the measurement.
            Gregorian calendar year, e.g., '1987'.
            If the value is not known, the value -1 should be entered as a
            dummy value.
        month (int): Month of the measurement.
            If the value is not known, the value -1 should be entered as a
            dummy value.
        day (int): Day of the measurement.
            If the value is not known, the value -1 should be entered as a
            dummy value.
        hour (int): Hour as part of the measurement time.
            In 24-hour clock format. If the value is not known, the value
            -1 should be entered as a dummy value.
        minute (int): Minute as part of the measurement time.
            If the value is not known, the value -1 should be entered as a
            dummy value.
        second (int): Second as part of the measurement time.
            If the value is not known, the value -1 should be entered as a
            dummy value.
        num_hours_advance_gmt (float): Number of hours in advance of Greenwich
            Mean Time.
        num_lines_block_comment (int): Number of lines of block comments.
        block_comment (str): Concatenated block comment lines.
            The number of occurrences of comment line is specified by the
            value of number of lines in block comment above.
        technique (str): Measurement Technique.
            One of the values 'AES diff' | 'AES dir' | 'EDX' | 'ELS' | 'FABMS'
            | 'FABMS energy spec' | 'ISS' | 'SIMS' | 'SIMS energy spec'
            | 'SNMS' | 'SNMS energy spec' | 'UPS' | 'XPS' | 'XRF'.
        x_coord (int): X-coordinate.
            The ordinal number, starting with unity, of the point in the array
            along the analysis source deflection system x-axis
            Only inserted if and only if the value of experiment mode is either
            **MAP** or **MAPDI**.
        y_coord (int): Y-coordinate.
            The ordinal number, starting with unity, of the point in the array
            along the analysis source deflection system y-axis.
            Only inserted if and only if the value of experiment mode is either
            **MAP** or **MAPDI**.
        values_exp_var (list): Values of experimental variables.
            May be, for example, total time in seconds, total etch time in
            seconds, temperature in Kelvin, energy in electron volts or mass
            in unified atomic mass units.
            The number of items in the list is specified by the value of
            :attr:`VamasBlock.num_experiment_variables
            <vamas.vamas_header.VamasHeader.num_experiment_variables>`,
            and the order in which the values are given is the same as the
            order in which :attr:`ExperimentVariable.label
            <vamas.vamas_header.ExperimentVariable.label>` and
            :attr:`ExperimentVariable.unit
            <vamas.vamas_header.ExperimentVariable.unit>` are declared.
        analysis_source_label (str): Label of the Analysis source.
        sputtering_z (int): Species of the sputtering ion or atom in atomic
            number.
            Inserted if and only if either (1) the value of *experiment mode*
            is **MAPDP**, **MAPSVDP**, **SDP** or **SDPSV**, or (2) the value
            of *technique* is **FABMS**, **FABMS energy spec**, **ISS**,
            **SIMS**, **SIMS energy spec**, **SNMS** or **SNMS energy spec**.
        sputtering_num_particles (float): Number of particles in sputtering ion
            or atom particle.
            Inserted if and only if either (1) the value of *experiment mode*
            is **MAPDP**, **MAPSVDP**, **SDP** or **SDPSV**, or (2) the value
            of *technique* is **FABMS**, **FABMS energy spec**, **ISS**,
            **SIMS**, **SIMS energy spec**, **SNMS** or **SNMS energy spec**.
        sputtering_charge (float): Charge sign and number of the sputtering
            ion or atom.
            Inserted if and only if either (1) the value of *experiment mode*
            is **MAPDP**, **MAPSVDP**, **SDP** or **SDPSV**, or (2) the value
            of *technique* is **FABMS**, **FABMS energy spec**, **ISS**,
            **SIMS**, **SIMS energy spec**, **SNMS** or **SNMS energy spec**.
        analysis_source_characteristic_energy (float): Characteristic energy
            of the analysis source in electron volts.
        analysis_source_strength (float): Strengh of the analysis source.
            Power of the analysis source in watts for XPS and XRF, beam current
            in nanoamps for AES, EDX, ISS, SIMS and SNMS; beam equivalent for
            FABMS.
        analysis_source_beam_width_x (float): Width of the beam in
            x-direction in micrometres at the sample in the plane perpendicular
            to the source beam.
        analysis_source_beam_width_y (float): Width of the beam in
            y-direction in micrometres at the sample in the plane perpendicular
            to the source beam.
        field_view_x (float): Field of view in x-direction in micrometres.
            Inserted if and only if the value of experiment mode is **MAP**,
            **MAPDP**, **MAPSV**, **MAPSVDP** or **SEM**.
        field_view_y (float) Field of view in y-direction in micrometres
            Inserted if and only if the value of experiment mode is **MAP**,
            **MAPDP**, **MAPSV**, **MAPSVDP** or **SEM**.
        linescan_coordinates (LinescanCoordinates): Linescan coordinates.
        analysis_source_polar_incidence_angle (float): Incidence angle of
            the analysis source.
            In degrees from upward z-direction, defined by the sample stage.
        analysis_source_azimuth (float): Azimuth angle of the analysis source.
            In degrees clockwise from the y-direction towards the operator,
            defined by the sample stage.
        analyzer_mode (str): Analyzer mode.
            The value is on of  'FAT' | 'FRR' | 'constant delta m'
            | 'constant m / delta m'.
        analyzer_pass_energy_or_retard_ratio_or_mass_res (float): Pass energy
            or retard ration or mass resolution of the analyser.
            The unit for energy is in electron volts, the one for mass in amu.
        differential_width (float): Differential width.
            Electron volts peak-to-peak for sinusoidal modulation or computer
            differentiation.
            Inserted if and only if the value of technique is **AES diff**.
        magnification_analyzer_transfer_lens (float): Magnification of the
            analyzer transfer lens.
        analyzer_work_function_or_acceptance_energy (float): Work function or
            acceptance angle of the analyzer.
            Positive value for work function in electron volts for AES, ELS,
            ISS, UPS and XPS. The acceptance energy of an ion is the energy
            filter pass energy of the mass spectrometer for FABMS, SIMS and
            SNMS.
        target_bias (float): Bias of the target.
            In volts, including the sign.
        analysis_width_x (float): Analysis width in x-direction.
            The analysis width x is the gated signal width of the source in the
            x-direction in the plane perpendicular to the beam for **FABMS**,
            **FABMS energy spec**, **ISS**, **SIMS**, **SIMS energy spec**,
            **SNMS** and **SNMS energy spec**, the analyser slit length divided
            by the magnification of the analyser transfer lens to that slit for
            **AES diff**, **AES dir**, **ELS**, **UPS** and **XPS**, and is the
            source width in the x-direction for both **EDX** and **XRF**.\n
            In micrometres.
        analysis_width_y (float): Analysis width in y-direction.
            The analysis width y is the gated signal width of the source in the
            y-direction in the plane perpendicular to the beam for **FABMS**,
            **FABMS energy spec**, **ISS**, **SIMS**, **SIMS energy spec**,
            **SNMS** and **SNMS energy spec**, the analyser slit length divided
            by the magnification of the analyser transfer lens to that slit for
            **AES diff**, **AES dir**, **ELS**, **UPS** and **yPS**, and is the
            source width in the y-direction for both **EDX** and **XRF**.\n
            In micrometres.
        analyzer_axis_take_off_polar_angle (float): Analyzer axis take off
            polar angle.
            Degrees from upward z-direction, defined by the sample stage.
        analyzer_axis_take_off_azimuth (float): Analyzer axis take off
            azimuth.
            Degrees clockwise from the y-direction towards the operator,
            defined by the sample stage.
        species_label (str): Label of the species.
            Elemental symbol or molecular formula.
        transition_or_charge_state_label (str): Label of the transition or
            charge state.
            For example, 'KLL' for **AES**, '1s' for **XPS**, '-1' for
            **SIMS**.
        charge_detected_particle (int): Charge of the detected particle.
            For example, -1 for **AES** and **XPS**, +1 for positive **SIMS**.
        x_label (str): Label of the x-values.
            Inserted if and only if the value of scan mode is **REGULAR**.
        x_units (str): Unit of the x-values.

            +------------+----------------------------------------------+
            | experiment | unit corresponding to technique              |
            |            |                                              |
            | mode       |                                              |
            |            +------------+------------+--------------------+
            |            | AES diff,  | FABMS,     | FABMS energy spec, |
            |            |            |            |                    |
            |            | AES dir,   | SIMS,      | SIMS energy spec,  |
            |            |            |            |                    |
            |            | EDX, ELSS, | SNMS,      | SNMS energy spec,  |
            |            |            |            |                    |
            |            | ISS, UPS   |            |                    |
            |            |            |            |                    |
            |            | XPS, XRF   |            |                    |
            +------------+------------+------------+--------------------+
            | MAP        | 'eV'       | 'u' or 's' | 'eV'               |
            |            |            |            |                    |
            | MAPDP      |            |            |                    |
            |            |            |            |                    |
            | NORM       |            |            |                    |
            |            |            |            |                    |
            | SDP        |            |            |                    |
            +------------+------------+------------+--------------------+
            | SDPSV      | 's'        | 's'        |                    |
            +------------+------------+------------+--------------------+

            Inserted if and only if the value of scan mode is **REGULAR**.

        x_start (float): Start value of the x-values.
            Inserted if and only if the value of scan mode is **REGULAR**.

        x_step (float): Step size between x-values.
            Inserted if and only if the value of scan mode is **REGULAR**.

        num_corresponding_variables (int): Number of corresponding variables.
            If the data is in the form of sets of corresponding values of two
            or more variables then
            :attr:`~VamasBlock.num_corresponding_variables` is equal to the
            number of variables, otherwise it is equal to unity.
        corresponding_variables (List[CorrespondingVariable]): Corresponding
            Variables.
        signal_mode (str): Signal mode.
            Is either 'analogue' | 'pulse counting'.
            Analogue signals, while recorded digitally, may be of either sign
            and have a gain which may be noted in the block comment.
            Pulse counting signals are integers with values equal to or greater
            than zero.
        signal_collection_time (float): Signal collection time.
            Time in seconds per scan for each channel or array-point, except
            for both **EDX** and **XRF** where it is the total spectrum
            collection time.
        num_scans_to_compile_block (int): Number of scan to compile block.
            This is the number of measurement recordings to get the data block,
            e.g. the number of sweeps for **XPS**.
        signal_time_correction (float): Time correction for the signal.
            This is the system dead time, except for **EDX** and **XRF** where
            it is the livetime-corrected acquisition time. In the case of a
            dead time, a positive value indicates that the count rate should be
            corrected by dividing by (1 - measured rate * dead time) whereas a
            negative value indicates a correction by multiplying
            by (exp(true count rate * dead time)). In seconds.
        sputtering_source (SputteringSource): Sputtering source.
        sample_normal_polar_angle_tilt (float): Sample normal polar angle of
            tilt.
            Degrees from upward z-direction, defined by the sample stage.
        sample_normal_tilt_azimuth (float): Sample normal tilt azimuth.
            Degrees clockwise from the y-direction towards the operator,
            defined by the sample stage.
        sample_rotation_angle (float): Sample rotation angle.
            Degrees clockwise rotation about the sample normal. If this is
            referenced to a particular direction on the sample this direction
            would be specified in a comment line at
            :attr:`~VamasBlock.block_comment`.
        num_additional_numerical_params (int): Number of additional numberical
            parameters.
        additional_numerical_params (List[AdditionalNumericalParam]):
            Additional numberical parameters.
        num_y_values (int): Number of y-values.
            The value of number of y-values (ordinate values) is equal to
            product of the value of
            :attr:`~VamasBlock.num_corresponding_variables` and
            the number of sets of :class:`CorrespondingVariable`s to be
            transferred.

    """

    block_identifier: str
    sample_identifier: str

    year: int
    month: int
    day: int
    hour: int
    minute: int
    second: int
    num_hours_advance_gmt: float

    num_lines_block_comment: int
    block_comment: str

    technique: str

    values_exp_var: List

    analysis_source_label: str

    analysis_source_characteristic_energy: float
    analysis_source_strength: float
    analysis_source_beam_width_x: float
    analysis_source_beam_width_y: float

    analysis_source_polar_incidence_angle: float
    analysis_source_azimuth: float

    analyzer_mode: str

    analyzer_pass_energy_or_retard_ratio_or_mass_res: float

    magnification_analyzer_transfer_lens: float
    analyzer_work_function_or_acceptance_energy: float

    target_bias: float

    analysis_width_x: float
    analysis_width_y: float
    analyzer_axis_take_off_polar_angle: float
    analyzer_axis_take_off_azimuth: float

    species_label: str
    transition_or_charge_state_label: str

    charge_detected_particle: int

    x_label: str
    x_units: str
    x_start: float
    x_step: float

    num_corresponding_variables: int
    corresponding_variables: List[CorrespondingVariable]

    signal_mode: str
    signal_collection_time: float
    num_scans_to_compile_block: int
    signal_time_correction: float

    sample_normal_polar_angle_tilt: float
    sample_normal_tilt_azimuth: float
    sample_rotation_angle: float

    num_additional_numerical_params: int
    additional_numerical_params: List[AdditionalNumericalParam]

    num_y_values: int

    # Optional
    x_coord: Optional[int] = None
    y_coord: Optional[int] = None
    sputtering_z: Optional[int] = None
    sputtering_num_particles: Optional[float] = None
    sputtering_charge: Optional[float] = None
    field_view_x: Optional[float] = None
    field_view_y: Optional[float] = None
    linescan_coordinates: Optional[LinescanCoordinates] = None
    differential_width: Optional[float] = None
    sputtering_source: Optional[SputteringSource] = None



class FileExtensionError(Exception):
    def __init__(self) -> None:
        message = (
            "The file is not a vamas file, file extension needs to be '.vms'"
        )
        super().__init__(message)


class VmsIdentifierError(Exception):
    def __init__(self) -> None:
        message = "The file does not contain the correct vamas identifier"
        super().__init__(message)
